#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <vector>
#include <string>
#include <map>
#include <set>

#include "acceptor.h"
#include "markpool.h"
#include "metacube.h"
#include "parse.h"
#include "server.h"
#include "serverpool.h"
#include "input.h"
#include "stats.h"
#include "state.pb.h"

using namespace std;

ServerPool *servers = NULL;
volatile bool hupped = false;

void hup(int ignored)
{
	hupped = true;
}

// Serialize the given state to a file descriptor, and return the (still open)
// descriptor.
int make_tempfile(const CubemapStateProto &state)
{
	char tmpl[] = "/tmp/cubemapstate.XXXXXX";
	int state_fd = mkstemp(tmpl);
	if (state_fd == -1) {
		perror("mkstemp");
		exit(1);
	}

	string serialized;
	state.SerializeToString(&serialized);

	const char *ptr = serialized.data();
	size_t to_write = serialized.size();
	while (to_write > 0) {
		ssize_t ret = write(state_fd, ptr, to_write);
		if (ret == -1) {
			perror("write");
			exit(1);
		}

		ptr += ret;
		to_write -= ret;
	}

	return state_fd;
}

// Read the state back from the file descriptor made by make_tempfile,
// and close it.
CubemapStateProto read_tempfile(int state_fd)
{
	if (lseek(state_fd, 0, SEEK_SET) == -1) {
		perror("lseek");
		exit(1);
	}

	string serialized;
	char buf[4096];
	for ( ;; ) {
		ssize_t ret = read(state_fd, buf, sizeof(buf));
		if (ret == -1) {
			perror("read");
			exit(1);
		}
		if (ret == 0) {
			// EOF.
			break;
		}

		serialized.append(string(buf, buf + ret));
	}

	close(state_fd);  // Implicitly deletes the file.

	CubemapStateProto state;
	if (!state.ParseFromString(serialized)) {
		fprintf(stderr, "PANIC: Failed deserialization of state.\n");
		exit(1);
	}

	return state;
}
	
// Reuse mark pools if one already exists.
MarkPool *get_mark_pool(map<pair<int, int>, MarkPool *> *mark_pools, int from, int to)
{
	pair<int, int> mark_range(from, to);
	if (mark_pools->count(mark_range) != 0) {
		return (*mark_pools)[mark_range];
	}

	// Check if we're overlapping some other mark pool.
	for (map<pair<int, int>, MarkPool *>::const_iterator mp_it = mark_pools->begin();
	     mp_it != mark_pools->end();
	     ++mp_it) {
		int other_from = mp_it->first.first;
		int other_to = mp_it->first.second;
		if ((from >= other_from && from < other_to) ||
		    (to >= other_from && to < other_to)) {
			fprintf(stderr, "WARNING: Mark pool %d-%d partially overlaps with %d-%d, you may get duplicate marks.\n",
				from, to, other_from, other_to);
			fprintf(stderr, "         Mark pools must either be completely disjunct, or completely overlapping.\n");
		}
	}	

	MarkPool *mark_pool = new MarkPool(from, to);
	mark_pools->insert(make_pair(mark_range, mark_pool));
	return mark_pool;
}
			
MarkPool *parse_mark_pool(map<pair<int, int>, MarkPool *> *mark_pools, const string &mark_str)
{
	size_t split = mark_str.find_first_of('-');
	if (split == string::npos) {
		fprintf(stderr, "WARNING: Invalid mark specification '%s' (expected 'X-Y'), ignoring.\n",
			mark_str.c_str());
		return NULL;
	}

	string from_str(mark_str.begin(), mark_str.begin() + split);
	string to_str(mark_str.begin() + split + 1, mark_str.end());
	int from = atoi(from_str.c_str());
	int to = atoi(to_str.c_str());

	if (from <= 0 || from >= 65536 || to <= 0 || to >= 65536) {
		fprintf(stderr, "WARNING: Mark pool range %d-%d is outside legal range [1,65536>, ignoring.\n",
			from, to);
		return NULL;
	}

	return get_mark_pool(mark_pools, from, to);
}
	
// Find all streams in the configuration file, and create inputs for them.
vector<Input *> create_inputs(const vector<ConfigLine> &config,
                              map<string, Input *> *deserialized_inputs)
{
	vector<Input *> inputs;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword != "stream") {
			continue;
		}
		assert(config[i].arguments.size() == 1);
		string stream_id = config[i].arguments[0];

		map<string, string>::const_iterator src_it =
			config[i].parameters.find("src");
		if (src_it == config[i].parameters.end()) {
			fprintf(stderr, "WARNING: stream '%s' has no src= attribute, clients will not get any data.\n",
				stream_id.c_str());
			continue;
		}

		string src = src_it->second;
		Input *input = NULL;
		map<string, Input *>::iterator deserialized_input_it =
			deserialized_inputs->find(stream_id);
		if (deserialized_input_it != deserialized_inputs->end()) {
			input = deserialized_input_it->second;
			if (input->get_url() != src) {
				fprintf(stderr, "INFO: Stream '%s' has changed URL from '%s' to '%s', restarting input.\n",
					stream_id.c_str(), input->get_url().c_str(), src.c_str());
				delete input;
				input = NULL;
			}
			deserialized_inputs->erase(deserialized_input_it);
		}
		if (input == NULL) {
			input = new Input(stream_id, src);
		}
		input->run();
		inputs.push_back(input);
	}
	return inputs;
}

int main(int argc, char **argv)
{
	fprintf(stderr, "\nCubemap starting.\n");

	struct timeval serialize_start;
	bool is_reexec = false;

	string config_filename = (argc == 1) ? "cubemap.config" : argv[1];
	vector<ConfigLine> config = parse_config(config_filename);

	int port = fetch_config_int(config, "port", 1, 65535, PARAMATER_MANDATORY);
	int num_servers = fetch_config_int(config, "num_servers", 1, 20000, PARAMATER_MANDATORY);  // Insanely high max limit.

	servers = new ServerPool(num_servers);

	CubemapStateProto loaded_state;
	int server_sock = -1, old_port = -1;
	set<string> deserialized_stream_ids;
	map<string, Input *> deserialized_inputs;
	if (argc == 4 && strcmp(argv[2], "-state") == 0) {
		is_reexec = true;

		fprintf(stderr, "Deserializing state from previous process... ");
		int state_fd = atoi(argv[3]);
		loaded_state = read_tempfile(state_fd);

		serialize_start.tv_sec = loaded_state.serialize_start_sec();
		serialize_start.tv_usec = loaded_state.serialize_start_usec();

		// Deserialize the streams.
		for (int i = 0; i < loaded_state.streams_size(); ++i) {
			servers->add_stream_from_serialized(loaded_state.streams(i));
			deserialized_stream_ids.insert(loaded_state.streams(i).stream_id());
		}

		// Deserialize the inputs. Note that we don't actually add them to any state yet.
		for (int i = 0; i < loaded_state.inputs_size(); ++i) {
			deserialized_inputs.insert(make_pair(
				loaded_state.inputs(i).stream_id(),
				new Input(loaded_state.inputs(i))));
		} 

		// Deserialize the server socket.
		server_sock = loaded_state.server_sock();
		old_port = loaded_state.port();

		fprintf(stderr, "done.\n");
	}

	// Find all streams in the configuration file, and create them.
	set<string> expecting_stream_ids = deserialized_stream_ids;
	map<pair<int, int>, MarkPool *> mark_pools;
	for (unsigned i = 0; i < config.size(); ++i) {
		if (config[i].keyword != "stream") {
			continue;
		}
		if (config[i].arguments.size() != 1) {
			fprintf(stderr, "ERROR: 'stream' takes exactly one argument\n");
			exit(1);
		}
		string stream_id = config[i].arguments[0];
		if (deserialized_stream_ids.count(stream_id) == 0) {
			servers->add_stream(stream_id);
		}
		expecting_stream_ids.erase(stream_id);

		// Set up marks, if so desired.
		if (config[i].parameters.count("mark")) {
			MarkPool *mark_pool = parse_mark_pool(&mark_pools, config[i].parameters["mark"]);
			servers->set_mark_pool(stream_id, mark_pool);
		}
	}

	// Warn about any servers we've lost.
	// TODO: Make an option (delete=yes?) to actually shut down streams.
	for (set<string>::const_iterator stream_it = expecting_stream_ids.begin();
	     stream_it != expecting_stream_ids.end();
	     ++stream_it) {
		string stream_id = *stream_it;
		fprintf(stderr, "WARNING: stream '%s' disappeared from the configuration file.\n",
			stream_id.c_str());
		fprintf(stderr, "         It will not be deleted, but clients will not get any new inputs.\n");
		if (deserialized_inputs.count(stream_id) != 0) {
			delete deserialized_inputs[stream_id];
			deserialized_inputs.erase(stream_id);
		}
	}

	// Open a new server socket if we do not already have one, or if we changed ports.
	if (server_sock != -1 && port != old_port) {
		fprintf(stderr, "NOTE: Port changed from %d to %d; opening new socket.\n", old_port, port);
		close(server_sock);
		server_sock = -1;
	}
	if (server_sock == -1) {
		server_sock = create_server_socket(port);
	}

	// See if the user wants stats.
	string stats_file = fetch_config_string(config, "stats_file", PARAMETER_OPTIONAL);
	int stats_interval = fetch_config_int(config, "stats_interval", 1, INT_MAX, PARAMETER_OPTIONAL, -1);
	if (stats_interval != -1 && stats_file.empty()) {
		fprintf(stderr, "WARNING: 'stats_interval' given, but no 'stats_file'. No statistics will be written.\n");
	}

	servers->run();

	AcceptorThread acceptor_thread(server_sock);
	acceptor_thread.run();

	vector<Input *> inputs = create_inputs(config, &deserialized_inputs);
	
	// All deserialized inputs should now have been taken care of, one way or the other.
	assert(deserialized_inputs.empty());
	
	if (is_reexec) {	
		// Put back the existing clients. It doesn't matter which server we
		// allocate them to, so just do round-robin. However, we need to add
		// them after the mark pools have been set up.
		for (int i = 0; i < loaded_state.clients_size(); ++i) {
			servers->add_client_from_serialized(loaded_state.clients(i));
		}
	}

	// Start writing statistics.
	StatsThread *stats_thread = NULL;
	if (!stats_file.empty()) {
		stats_thread = new StatsThread(stats_file, stats_interval);
		stats_thread->run();
	}

	signal(SIGHUP, hup);
	
	struct timeval server_start;
	gettimeofday(&server_start, NULL);
	if (is_reexec) {
		// Measure time from we started deserializing (below) to now, when basically everything
		// is up and running. This is, in other words, a conservative estimate of how long our
		// “glitch” period was, not counting of course reconnects if the configuration changed.
		double glitch_time = server_start.tv_sec - serialize_start.tv_sec +
			1e-6 * (server_start.tv_usec - serialize_start.tv_usec);
		fprintf(stderr, "Re-exec happened in approx. %.0f ms.\n", glitch_time * 1000.0);
	}

	while (!hupped) {
		usleep(100000);
	}

	// OK, we've been HUPed. Time to shut down everything, serialize, and re-exec.
	gettimeofday(&serialize_start, NULL);

	if (stats_thread != NULL) {
		stats_thread->stop();
	}
	acceptor_thread.stop();

	CubemapStateProto state;
	state.set_serialize_start_sec(serialize_start.tv_sec);
	state.set_serialize_start_usec(serialize_start.tv_usec);
	state.set_server_sock(server_sock);
	state.set_port(port);

	for (size_t i = 0; i < inputs.size(); ++i) {
		inputs[i]->stop();
		state.add_inputs()->MergeFrom(inputs[i]->serialize());
	}

	for (int i = 0; i < num_servers; ++i) {	
		servers->get_server(i)->stop();

		CubemapStateProto local_state = servers->get_server(i)->serialize();

		// The stream state should be identical between the servers, so we only store it once.
		if (i == 0) {
			state.mutable_streams()->MergeFrom(local_state.streams());
		}
		for (int j = 0; j < local_state.clients_size(); ++j) {
			state.add_clients()->MergeFrom(local_state.clients(j));
		}
	}
	delete servers;

	fprintf(stderr, "Serializing state and re-execing...\n");
	int state_fd = make_tempfile(state);

	char buf[16];
	sprintf(buf, "%d", state_fd);

	for ( ;; ) {
		execlp(argv[0], argv[0], config_filename.c_str(), "-state", buf, NULL);
		perror("execlp");
		fprintf(stderr, "PANIC: re-exec of %s failed. Waiting 0.2 seconds and trying again...\n", argv[0]);
		usleep(200000);
	}
}
